# -*- coding:utf-8 -*-
"""Module of simple local proxy.

This module defines the following classes:

class HTTPRequest()
    HTTPRequest is a class of HTTP request.
    It stores HTTP request headers and content.
    It also provides parse_header(header_lines) method.

class HTTPResponse()
    HTTPResponse is a class of HTTP response.
    It stores HTTP response headers and content.
    It also provides parse_header(headers) method.

class ProxyConnection(client_socket)
    ProxyConnection is class for a connection via local proxy.
    This class provides basic proxy functionality.

Here is an example of simple proxy sever script utilizing proxylib module.

[Example]
    import socket
    import proxylib
    
    # Read APID
    apid = socket.select_access_point()
    ap = socket.access_point(apid)
    socket.set_default_access_point(ap)
    
    # Create server socket
    server_socket = socket.socket(socket.AF_INET,
                                  socket.SOCK_STREAM)
    server_socket.bind(('127.0.0.1', 8080))
    server_socket.listen(5)
    
    def wait():
        # Wait for connection
        server_socket.accept(cb)
    
    def cb((sc, addr)):
        # Callback function
        wait()
        client_socket = socket._socketobject(sc, socket.AF_INET)
        conn = proxylib.ProxyConnection(client_socket)
        conn.read_request()
        conn.get_response()
        conn.close()
    
    wait()
"""

import e32
_version14 = 1.4
_version20 = 2.0
if e32.pys60_version.startswith('1.4'):
    pys60_version = _version14
elif e32.pys60_version.startswith('1.9'):
    pys60_version = _version20
else:
    pys60_version = None

import socket
if pys60_version == _version14:
    import select
import httplib
from urlparse import urlsplit

__all__ = ['ProxyException', 'InvalidSocket', 'InvalidRequest',
           'CannotReadRequest', 'CannotSendRequest', 'InvalidResponse',
           'CannotReadResponse', 'CannotSendResponse', 'HTTPRequest',
           'HTTPResponse', 'ProxyConnection']

__author__ = 'Hiisi (hiisi.proxy@gmail.com)'
__date__ = '2009-01-01'
__version__ = '0.0.1'
__credits__ = 'Guido van Rossum, for an excellent programming language.'

# Proxy status
_STATE_IDLE = 'Idle'
_STATE_REQUESTREAD = 'Request-read'
_STATE_REQUESTSENT = 'Request-sent'
_STATE_RESPONSEREAD = 'Response-read'
_STATE_RESPONSESENT = 'Response-sent'
_STATE_ERROR = 'Error'

# Server Configuration
_MAX_BYTE = 300000          # Maximum bytes for 1 communication
_BUFSIZE = 4096             # Buffer size (byte)
_PORT_HTTP = 80             # Port for HTTP connection
_PORT_HTTPS = 443           # Port for HTTPS connection
_HTTP_TIMEOUT = 20          # Timeout for http connection (sec)
_PERIOD_RETRY = 0.2         # Time period before retry (sec)
_MAX_RETRY = _HTTP_TIMEOUT/_PERIOD_RETRY

# Protocols
_PROTOCOL_HTTP = ['HEAD', 'GET', 'POST']
_PROTOCOL_HTTPS = ['CONNECT']

# Exceptions
class ProxyException(Exception):
    pass

class InvalidSocket(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Invalid socket.')

class InvalidRequest(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Invalid request.')

class CannotReadRequest(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Cannot read request')

class CannotSendRequest(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Cannot send request')

class InvalidResponse(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Invalid response.')

class CannotReadResponse(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Cannot read response')

class CannotSendResponse(ProxyException):
    def __init__(self):
        pass
    def __str__(self):
        return repr(u'Cannot send response')

class HTTPRequest:
    """Class of HTTPRequest"""
    def __init__(self):
        self.method = ''
        self.url = ''
        self.host = ''
        self.port = ''
        self.path = ''
        self.protocol = ''
        self.header = {}
        self.content = ''
        self.content_length = 0
    
    def parse_header(self, header_lines):
        """Parse HTTP request header"""
        if not isinstance(header_lines, list):
            raise InvalidRequest
        try:
            # Parse first line
            method, url, protocol = \
             header_lines.pop(0).split(' ', 2)
            
            self.method = method
            self.url = url
            self.protocol = protocol
            
            if self.method == 'CONNECT':
                schceme, netloc, path, query, frag = \
                 'https', self.url, self.url, '', ''
            else:
                schceme, netloc, path, query, frag = urlsplit(self.url)
            
            desc = netloc.find(':')
            if desc != -1:
                host = netloc[:desc]
                port = int(netloc[(desc+1):])
            else:
                host = netloc
                port = _PORT_HTTP
            self.host = host
            self.port = port
            self.path = path
            if query:
                self.path += '?%s' % query
            if frag:
                self.path += '#%s' % frag
            
            # Parse other lines
            for line in header_lines:
                desc = line.find(': ')
                if desc != -1:
                    self.header[line[:desc]] = line[(desc+2):]
                    if line[:desc].lower() == 'content-length':
                        self.content_length = int(self.header[line[:desc]])
        except ValueError:
            raise InvalidRequest

class HTTPResponse:
    """Class of HTTP Response"""
    def __init__(self):
        self.protocol = ''
        self.code = 0
        self.msg = ''
        self.header = {}
        self.content = ''
        self.content_length = 0
    
    def parse_header(self, header_lines):
        """Parse HTTP response header"""
        if not isinstance(header_lines, list):
            raise InvalidResponse
        
        try:
            # Parse first line
            protocol, code, msg = \
             header_lines.pop(0).split(' ', 2)
            
            self.protocol = protocol
            self.code = int(code)
            self.msg = msg
            
            # Parse other lines
            for line in header_lines:
                desc = line.find(': ')
                if desc != -1:
                    self.header[line[:desc]] = line[(desc+2):]
                    if line[:desc].lower() == 'content-length':
                        self.content_length = int(self.header[line[:desc]])
        except ValueError:
            raise InvalidResponse

class ProxyConnection:
    """Class for proxy connection"""
    _userange = 0
    _connection = httplib.HTTPConnection
    _request = HTTPRequest
    _response = HTTPResponse
    
    def __init__(self, client_socket):
        """Initialize proxy connection"""
        if not isinstance(client_socket, socket._socketobject):
            self.state = _STATE_ERROR
            raise InvalidSocket
        
        self.client_socket = client_socket
        self.req = self._request()
        self.res = self._response()
        self.state = _STATE_IDLE
    
    def read_request(self):
        """Receive request from client socket"""
        if self.state != _STATE_IDLE:
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotReadRequest
        try:
            header_lines = []
            fp = self.client_socket.makefile('rb')
            # Read HTTP headers
            while 1:
                buf = fp.readline().rstrip()
                if not buf:
                    break
                else:
                    header_lines.append(buf)
            
            if header_lines:
                # Parse header
                self.req.parse_header(header_lines)
                
                # Read request body (POST method)
                if self.req.content_length:
                    self.req.content = fp.read(self.req.content_length)
                
                fp.close()
                self.state = _STATE_REQUESTREAD
            else:
                fp.close()
                raise IOError
        
        except (IOError, InvalidRequest):
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotReadRequest
    
    def send_response(self):
        """Send response to client socket"""
        if not self.state in [_STATE_RESPONSEREAD, _STATE_ERROR]:
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotSendResponse
        
        response = '%s %d %s\r\n' % (self.res.protocol, self.res.code, \
                                     self.res.msg)
        for key in self.res.header.keys():
            response += '%s: %s\r\n' % (key, self.res.header[key])
        response += '\r\n'
        response += self.res.content
        try:
            self.client_socket.send(response)
        except:
            raise CannotSendResponse
    
    def get_response(self, proxyhost=None, proxyport=None):
        """Redirect to HTTP_connect or HTTPS_connect or not_implemented"""
        if self.req.method in _PROTOCOL_HTTP:
            self.HTTP_session(proxyhost, proxyport)
        elif conn.req.method in _PROTOCOL_HTTPS and \
             pys60_version == _version14:
            self.HTTPS_session(proxyhost, proxyport)
        else:
            self.not_implemented()
            self.send_response()
    
    def HTTP_session(self, proxyhost=None, proxyport=None):
        """Send request and receive response"""
        if not isinstance(proxyhost, str):
            proxyhost = None
        if not isinstance(proxyport, int):
            proxyport = None
        
        self.HTTP_connect(proxyhost, proxyport)
        self.send_response()
    
    def HTTP_connect(self, proxyhost=None, proxyport=None):
        """Send request and receive response"""
        # Set default access point
        if self.state != _STATE_REQUESTREAD:
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotSendRequest
        
        if not isinstance(proxyhost, str):
            proxyhost = None
        if not isinstance(proxyport, int):
            proxyport = None
        if proxyhost is not None:
            host = proxyhost
        else:
            host = self.req.host
        if proxyport is not None:
            port = proxyport
        else:
            port = _PORT_HTTP
            
        try:
            # Connect
            conn = self._connection(host, port)
            byteread = 0
            while 1:
                if self._userange:
                    self.req.header['Range'] = \
                     'bytes=%d-%d' % (byteread, (byteread + _MAX_BYTE - 1))
                
                # Send Request
                conn.request(self.req.method, self.req.url, self.req.content, \
                             self.req.header)
                self.state = _STATE_REQUESTSENT
                
                # Get response
                res = conn.getresponse()
                
                # Read header
                if res.version == 9:
                    protocol = 'HTTP/0.9'
                elif res.version == 10:
                    protocol = 'HTTP/1.0'
                else:
                    protocol = 'HTTP/1.1'
                header_lines = ['%s %d %s' % (protocol, res.status, \
                                              res.reason)]
                for header_line in res.msg.headers:
                    header_lines.append(header_line.rstrip())
                self.res.parse_header(header_lines)
                
                # Read content
                buf = res.read()
                byteread += len(buf)
                self.res.content += buf
                self.res.content_length = len(self.res.content)
                
                for key in self.res.header.keys():
                    if key.lower() == 'content-length':
                        del self.res.header[key]
                    elif key.lower() == 'transfer-encoding':
                        del self.res.header[key]
                self.res.header['Content-Length'] = str(self.res.content_length)
                
                if self.res.code == 206:
                    if buf:
                        continue
                    else:
                        self.res.code = 200
                        self.res.msg = 'OK'
                        break
                else:
                    break
            conn.close()
            self.state = _STATE_RESPONSEREAD
        except socket.gaierror:
            conn.close()
            self.set_error(504, 'Gateway Timeout')
            self.state = _STATE_ERROR
            raise CannotSendRequest
        except (IOError, InvalidRequest):
            conn.close()
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotReadResponse
        except:
            conn.close()
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotReadResponse
    
    def HTTPS_session(self, proxyhost=None, proxyport=None):
        """HTTPS tunneling and return response"""
        if self.state != _STATE_REQUESTREAD:
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotSendRequest
        
        if not isinstance(proxyhost, str):
            proxyhost = None
        if not isinstance(proxyport, int):
            proxyport = None
        if proxyport is not None:
            port = proxyport
        else:
            port = _PORT_HTTPS
        
        try:
            # Create HTTP socket
            HTTP_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            HTTP_socket.setblocking(0)
            
            if proxyhost is None:
                host = self.req.host
                # Connect to SSL server
                HTTP_socket.connect((host, port))
                
                # Return response
                self.res.protocol = 'HTTP/1.1'
                self.res.code = 200
                self.res.msg = 'Connection Established'
                self.res.header = {'Content-Length': '0'}
                self.res.content = ''
                self.res.content_length = 0
                
                self.state = _STATE_RESPONSEREAD
            else:
                host = proxyhost
                # Delete Connection header
                for key in self.req.header.keys():
                    if key.lower() == 'Connection':
                        del self.req.header[key]
                        break
                # Connect to proxy server
                HTTP_socket.connect((host, port))
                
                # Send requeset
                request = '%s %s %s\r\n' % (self.req.method, self.req.url, \
                                            self.req.protocol)
                for key in self.req.header.keys():
                    request += '%s: %s\r\n' % (key, self.req.header[key])
                
                request += '\r\n'
                request += self.req.content
                HTTP_socket.send(request)
                self.state = _STATE_REQUESTSENT
                
                # Receive response
                firstline = ''
                isfirstline = 1
                headers = []
                fp = HTTP_socket.makefile('rb')
                while 1:
                    buf = fp.readline().rstrip()
                    if not buf:
                        break
                    
                    if isfirstline:
                        firstline = buf
                        isfirstline = 0
                    else:
                        headers.append(buf)
                
                if not isfirstline:
                    # Return response header
                    self.res.protocol, self.res.code, self.res.msg = \
                     firstline.split(' ', 2)
                    self.res.code = int(self.res.code)
                    self.res.content = ''
                    self.res.content_length = 0
                    
                    # Read content
                    for key in self.res.header.keys():
                        if key.lower() == 'content-length':
                            self.res.content_length = int(self.res.header[key])
                            break
                    self.res.content = fp.read(self.res.content_length)
                    self.state = _STATE_RESPONSEREAD
                else:
                    fp.close()
                    self.state = _STATE_ERROR
                    raise IOError
            
            self.send_response()
            self.state = _STATE_RESPONSESENT
        except socket.gaierror:
            HTTP_socket.close()
            self.set_error(504, 'Gateway Timeout')
            self.state = _STATE_ERROR
            raise CannotSendRequest
        except IOError:
            HTTP_socket.close()
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotSendRequest
        except:
            HTTP_socket.close()
            self.set_error(500, 'Internal Server Error')
            self.state = _STATE_ERROR
            raise CannotSendRequest
        
        try:
            # Start tunneling
            retry = _MAX_RETRY
            while retry > 0:
                readables, writables, exceptions = \
                 select.select([self.client_socket, HTTP_socket], \
                               [], [], _PERIOD_RETRY)
                if readables:
                    for readable in readables:
                        try:
                            buf = readable.recv(_BUFSIZE)
                        except:
                            retry -= 1
                            continue
                        if buf:
                            if readable == self.client_socket:
                                HTTP_socket.send(buf)
                            else:
                                self.client_socket.send(buf)
                        else:
                            retry -= 1
                else:
                    retry -= 1
            
            # Close HTTP socket
            HTTP_socket.close()
        except IOError:
            HTTP_socket.close()
        except:
            pass
    
    def not_implemented(self):
        """Not Implemented Error"""
        self.set_error(501, 'Not Implemented')
        self.state = _STATE_RESPONSEREAD
    
    def close(self):
        """Close connection"""
        self.state = _STATE_IDLE
    
    def set_error(self, code, msg):
        """Set error response"""
        if not isinstance(code, int):
            code = 500
        if not isinstance(msg, str):
            msg = 'Internal Server Error'
        self.res.protocol = 'HTTP/1.1'
        self.res.code = code
        self.res.msg = msg
        self.res.header = {'Content-Length': '0'}
        self.res.content = ''
        self.res.content_length = 0

def main():
    pass

if __name__ == '__main__':
    main()