#-*- coding:utf-8 -*-
"""Ease WAP communication with urllib module.

This tiny module will enables urllib module to:
- set WAP proxy.
- add arbitrary HTTP headers.
- change user agent string.

It defines the following public functions and a class:

wap_connection([headers])
    wap_connection is a function which makes connection via WAP gateway.
    wap_connection sets additional HTTP headers.

direct_connection([headers])
    direct_connection is a function which makes connection direct.
    direct_connection sets additional HTTP headers.

class WAPURLopener(proxy, headers)
    WAPURLopener is a subclass of urllib.FancyURLopener.
    This class provides HTTP header insertion functionality.

Here is an example script utilizing it.

[Example]
    import socket
    import urllib
    import waplib
    
    # Read APID
    apid = socket.select_access_point()
    ap = socket.access_point(apid)
    socket.set_default_access_point(ap)

    # Set connection mode
    if apid == waplib.APID_WAP:
        waplib.wap_connection({'Accept-Charset': 'utf-8'})
    else:
        waplib.direct_connection({})

    # Connect
    f = urllib.urlopen('http://www.python.org/')
    f.read()
    f.close()

In order to utilize it, each user must follow the Installation
Guide below in advance.

[Installation Guide]
    1. Edit APID_WAP, the access point ID for WAP Communication.
       Check APID in Python Script Shell (in Interactive mode).
       >>> import socket
       >>> print socket.access_points()
    2. Edit PROXY_HOST and PROXY_PORT, the host name and port number of WAP
       gateway, respectively.
    3. (Optional) Edit other constants.
    4. Save this module in E:\Python\lib\
"""

import httplib
import urllib

__all__ = ['WAPURLopener', 'wap_connection', 'direct_connection']

__author__ = 'Hiisi (hiisi.proxy@gmail.com)'
__date__ = '2008-08-22'
__version__ = '0.0.3'
__credits__ = 'Guido van Rossum, for an excellent programming language.'

# Parameters (users edit this part.)
APID_WAP = 1                                                      # APID
PROXY_HOST = ''                                                   # Proxy host
PROXY_PORT = 8080                                                 # Proxy port
PROXY_TYPE = 'http'                                               # Proxy type
PROXY = '%s://%s:%s/' % (PROXY_TYPE, PROXY_HOST, str(PROXY_PORT)) # Proxy URL
USER_AGENT = urllib.URLopener.version                             # User agent
WAP_PROFILE = ''                                                  # WAP profile

class WAPURLopener(urllib.FancyURLopener):
    """A new class overriding urllib.FancyURLopener."""
    def __init__(self, proxy, headers):
        """Create a WAPURLopener instance."""
        urllib.FancyURLopener.__init__(self, proxy)
        # Add HTTP headers.
        self.addheaders = []
        for k, v in headers.iteritems():
            self.addheaders.append((k, v))

def wap_connection(headers={}):
    """Set connection mode: via WAP"""
    # Set WAP proxy
    proxy = {PROXY_TYPE : PROXY}
    # Set missing HTTP headers (if needed).
    if not headers.has_key('User-Agent'):
        headers['User-Agent'] = USER_AGENT
    if WAP_PROFILE and not headers.has_key('x-wap-profile'):
        headers['x-wap-profile'] = WAP_PROFILE
    # Set URL opener: with proxy
    urllib._urlopener = WAPURLopener(proxy, headers)
    # Change HTTP version from 1.0 to 1.1.
    httplib.HTTP._http_vsn = 11
    httplib.HTTP._http_vsn_str = 'HTTP/1.1'

def direct_connection(headers={}):
    """Set connection mode: direct"""
    # Set missing HTTP headers (if needed).
    if not headers.has_key('User-Agent'):
        headers['User-Agent'] = USER_AGENT
    # Set URL opener: without proxy
    urllib._urlopener = WAPURLopener(None, headers)
    # Revert HTTP version from 1.1 to 1.0.
    httplib.HTTP._http_vsn = 10
    httplib.HTTP._http_vsn_str = 'HTTP/1.0'

# These are the obsolete function names.
setparams = wap_connection
resetparams = direct_connection

def main():
    pass

if __name__ == '__main__':
    main()
